/*
  $Id: fatresize.c,v 1.4 2005/04/13 15:51:50 siome_tajshe Exp $
  Copyright (C) 2005  Anton D. Kachalov <mouse@altlinux.org>

  The FAT16/FAT32 non-destructive resizer.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <getopt.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <parted/parted.h>

#include "config.h"

#define FAT32MIN	1024*1024*512

static struct {
    unsigned char *dev;
    unsigned char pnum;
    unsigned int size;
    char verbose;
    char progress;
    char is_evms;
    char info;
} opts;

static void usage()
{
    fprintf(stderr, "Usage: %s [options] device (e.g. /dev/hda1, /dev/sda2)\n"
		    "    Resize an FAT16/FAT32 volume non-destructively:\n\n"
		    "    -s, --size SIZE    Resize volume to SIZE[k|M|G] bytes\n"
		    "    -i, --info         Show volume information\n"
		    "    -p, --progress     Show progress\n"
		    "    -q, --quite        Be quite\n"
		    "    -v, --verbose      Verbose\n"
		    "    -h, --help         Display this help\n\n"
		    "Please report bugs to %s\n",
		    PACKAGE_NAME, PACKAGE_BUGREPORT);

    exit(1);
}

static void printd(int level, const char *fmt, ...)
{
    va_list ap;

    if (opts.verbose < level)
	return;

    va_start(ap, fmt);
    vprintf(fmt, ap);
    va_end(ap);
}

static int get_size(char *s)
{
    long int size;
    char *suffix;
    int prefix_kind = 1024;

    size = strtoll(s, &suffix, 10);
    if (size <= 0 || errno == ERANGE)
    {
	fprintf(stderr, "Illegal new volume size\n");
	usage();
    }

    if (!*suffix)
	return size;

    if (suffix[1])
	usage();

    switch (*suffix)
    {
	case 'G':
	    size *= prefix_kind;
	case 'M':
	    size *= prefix_kind;
	case 'k':
	    size *= prefix_kind;
	    break;
	default:
	    usage();
    }

    return size;
}

static int get_partnum(char *dev)
{
    int pnum;
    char *p;

    if (opts.is_evms)
	return 1;

    p = dev+strlen(dev)-1;
    while (*p && isdigit(*p) && *p != '/')
	p--;

    pnum = atoi(p+1);
    return pnum ? pnum : 1;
}

static char *get_devname(char *dev)
{
    char *devname;
    char *p;

    if (opts.is_evms)
	return dev;

    p = dev+strlen(dev)-1;
    while (*p && isdigit(*p) && *p != '/')
	p--;

    devname = malloc(p-dev+2);
    strncpy(devname, dev, p-dev+1);
    devname[p-dev+1] = '\0';

    return devname;
}

static void resize_handler(PedTimer *timer, void *ctx)
{
    printf(".");
    fflush(stdout);
}

static PedExceptionOption fatresize_handler(PedException *ex)
{
    if (opts.verbose != -1)
    {
	fprintf(stderr, "%s: %s\n", ped_exception_get_type_string(ex->type),
				    ex->message);
    }

    switch (ex->options)
    {
	case PED_EXCEPTION_IGNORE_CANCEL:
	    return PED_EXCEPTION_IGNORE;
	default:
	    return PED_EXCEPTION_UNHANDLED;
    }
}

int main(int argc, char **argv)
{
    int opt;

    PedDevice *dev;
    PedDisk *disk;
    PedPartition *part;
    PedTimer *timer = NULL;

    PedSector endM;
    PedFileSystem *fs;
    PedGeometry new_geom;
    PedConstraint *constraint;

    static const char *sopt = "-his:vpq";
    static const struct option lopt[] = {
	{"help",	no_argument,		NULL, 'h'},
	{"info",	no_argument,		NULL, 'i'},
	{"progress",	no_argument,		NULL, 'p'},
	{"size",	required_argument,	NULL, 's'},
	{"verbose",	no_argument,		NULL, 'v'},
	{"quite",	no_argument,		NULL, 'q'},
	{NULL, 0, NULL, 0}
    };

    memset(&opts, 0, sizeof(opts));

    if (argc < 2)
	usage();

    while ((opt = getopt_long(argc, argv, sopt, lopt, NULL)) != -1)
    {
	switch (opt)
	{
	    case 1:
		if (!opts.dev)
		{
		    if (!strncmp(optarg, "/dev/evms/", 10))
			opts.is_evms = 1;
		    opts.dev = get_devname(optarg);
		    opts.pnum = get_partnum(optarg);
		}
		else
		    usage();
		break;
	    case 'i':
		opts.info = 1;
		break;
	    case 'p':
		opts.progress = 1;
		break;
	    case 's':
		opts.size = get_size(optarg);
		break;
	    case 'v':
		opts.verbose++;
		break;
	    case 'q':
		opts.verbose = -1;
		break;
	    case 'h':
	    case '?':
	    default:
		printd(0, "%s (%s)\n", PACKAGE_STRING, BUILD_DATE);
		usage();
	}
    }

    printd(0, "%s (%s)\n", PACKAGE_STRING, BUILD_DATE);

    if (!opts.dev)
    {
	fprintf(stderr, "You must specify exactly one device.\n");
	return 1;
    }
    else if (!opts.size && !opts.info)
    {
	fprintf(stderr, "You must specify new size.\n");
	return 1;
    }

    ped_exception_set_handler(fatresize_handler);

    if (opts.progress)
	timer = ped_timer_new(resize_handler, NULL);

    dev = ped_device_get(opts.dev);
    if (!dev)
	return 1;
    disk = ped_disk_new(dev);
    if (!disk)
	return 1;
    part = ped_disk_get_partition(disk, opts.pnum);
    if (!part || !part->fs_type)
	return 1;

    if (strncmp(part->fs_type->name, "fat", 3))
    {
	if (opts.is_evms)
	{
	    ped_exception_throw(PED_EXCEPTION_ERROR, PED_EXCEPTION_CANCEL,
				"%s is not valid FAT16/FAT32 partition.",
				opts.dev);
	}
	else
	{
	    ped_exception_throw(PED_EXCEPTION_ERROR, PED_EXCEPTION_CANCEL,
				"%s%d is not valid FAT16/FAT32 partition.",
				opts.dev, opts.pnum);
	}
	return 1;
    }

    printd(3, "ped_file_system_open()\n");
    fs = ped_file_system_open(&part->geom);
    if (!fs)
	return 1;

    printd(3, "ped_file_system_get_resize_constraint()\n");
    constraint = ped_file_system_get_resize_constraint(fs);

    if (opts.info)
    {
	printf("FAT: %s\n", part->fs_type->name);
	printf("Size: %lld\n", part->geom.length * dev->sector_size);
	printf("Min size: %lld\n", constraint->min_size * dev->sector_size);
	printf("Max size: %lld\n", constraint->max_size * dev->sector_size);

	ped_constraint_destroy(constraint);
	return 0;
    }

    /* FAT32 must be bigger than 512Mb */
    if (part->fs_type->name[3] == '3' && opts.size < FAT32MIN)
    {
	ped_exception_throw(PED_EXCEPTION_ERROR, PED_EXCEPTION_CANCEL,
			    "FAT32 partition must be bigger than 512Mb.");
	return 1;
    }

    endM = opts.size / dev->sector_size;

    /* If we running with EVMS node, make it more smaller */
    if (opts.is_evms && endM > 0)
	endM--;

    if (endM > dev->length - 1 && endM < dev->length * 1.01)
    {
	/*
	 * FIXME: must be error?
	 *	  show human readable size
	 */
        endM = dev->length - 1;
	ped_exception_throw(PED_EXCEPTION_WARNING, PED_EXCEPTION_IGNORE,
			    "Too big partition. Setting size to %.2f Mb.",
			    (double)(endM * dev->sector_size) / (1024*1024));
    }

    if (endM < constraint->min_size)
    {
	/*
	 * FIXME: show human readable size
	 */
	ped_exception_throw(PED_EXCEPTION_WARNING, PED_EXCEPTION_IGNORE,
			    "Too small partition. Setting size to %.2f Mb.",
			    (double)(constraint->min_size * dev->sector_size) / (1024*1024));
    }

    printd(3, "ped_geometry_init()\n");
    if (!ped_geometry_init(&new_geom, dev, part->geom.start,
					    endM - part->geom.start + 1))
	return 1;

    printd(3, "ped_disk_set_partition_geom()\n");
    if (!ped_disk_set_partition_geom(disk, part, constraint,
				     new_geom.start, new_geom.end))
	return 1;

    printd(1, "Resizing partition.\n");
    if (!ped_file_system_resize(fs, &part->geom, timer))
	return 1;

    printd(1, "Done.\n");
    ped_partition_set_system(part, fs->type);
    ped_file_system_close(fs);

    ped_constraint_destroy(constraint);
    if (!opts.is_evms)
    {
	printd(1, "Committing changes.\n");
	if (!ped_disk_commit(disk))
	    return 1;
    }
    else
    {
	printd(1, "Commiting changes only to disk.\n");
	if (!ped_disk_commit_to_dev(disk))
	    return 1;
    }

    return 0;
}
